#! /usr/bin/env python

# CONSTANTS
MAX_LOCAL_DIST = 5000

#Class to handle split alignments from a single contig
class AnomalousContigCls:
  #alignment1 in pair
  align1 = None
  #alignment1 in pair
  align2 = None
  #contig number
  contig = None
  #bp overlap in contig alignments
  overlap_query = None
  #fraction of alignments that overlap in contig coordinates
  overlap_query_fraction = None
  #bp's covered by the two alignments
  contig_cover = None
  #fraction of total contig covered by two alingments
  contig_cover_fraction = None
  # type of event: chimeric, exon duplication, local inversion
  type = None

  def __init__(self, a1, a2):
    #Read in the alignment pairs (ordered by query start co-ordinate)
    if (a1.qstart <= a2.qstart):
      self.align1 = fix_align(a1)
      self.align2 = fix_align(a2)
    else: # a1.qstart > a2.qstart
      self.align1 = fix_align(a2)
      self.align2 = fix_align(a1)
    # end if
    #attributes of the contig
    self.contig = a1.query
    self.contig_size = a1.query_len
    # calculate the target distance
    self.target_distance = self.calc_target_distance()
    # set the type to unknown by default
    self.type = "unknown"
  # end def

  # attempt to determine the type of the event
  def determine_type(self):
    # check whether the alignments are to the same chromosome
    if(self.align1.target == self.align2.target):
      # check whether the alignments are to the same strand
      if(self.align1.query_strand == self.align2.query_strand):
        # check whether one target region is a subset of the other
        if (self.subset_target()):
          self.type = "end-duplication"
        # check whether the query alignment order is the opposite of the
        # target alignment order
        elif("forward" != self.get_target_order()):
          # check whether the target regions are too far apart to be
          # considered a duplication
          if (MAX_LOCAL_DIST < self.target_distance):
            self.type = "intrachr-non-colinear"
          else:
            # they might be an exon duplication event
            self.type = "junction-duplication"
          # end if
        else:
          if (MAX_LOCAL_DIST < self.target_distance):
            self.type = "intrachr-same-strand"
          else:
            self.type = "read-through"
          # end if
        # end if
      else:
        if (MAX_LOCAL_DIST < self.target_distance):
          self.type = "intrachr-opp-strand"
        else:
          # they might be a local inversion event
          self.type = "local-inversion"
        # end if
      # end if
    else:
      # they are some sort of interchromosomal chimeric event
      self.type = "interchr"
    # end if
  # end def

  def get_target_order(self):
    if (self.align1.target != self.align2.target or
        self.align1.query_strand != self.align2.query_strand):
      return "NA"
    # end if
    target_order = "equal"
    # use the strand to determine the order of the alignments in
    # target (genomic) co-ordinates
    if ("+" == self.align1.query_strand):
      if (self.align1.tstart < self.align2.tstart):
        target_order = "forward"
      elif (self.align1.tstart > self.align2.tstart):
        target_order = "reverse"
      # end if
    else: # "-" == self.align1.query_strand
      if (self.align1.tstart < self.align2.tstart):
        target_order = "reverse"
      elif (self.align1.tstart > self.align2.tstart):
        target_order = "forward"
      # end if
    # end if
    return target_order
  # end def

  def subset_target(self):
    if (self.align1.target != self.align2.target):
      return False
    # end if
    start1 = min(self.align1.tstart, self.align1.tend)
    end1   = max(self.align1.tstart, self.align1.tend)
    start2 = min(self.align2.tstart, self.align2.tend)
    end2   = max(self.align2.tstart, self.align2.tend)
    if ((start1 <= start2 and end1 >= end2) or
        (start2 <= start1 and end2 >= end1)):
      return True
    # end if
    return False
  # end def

  def calc_target_distance(self):
    if (self.align1.target != self.align2.target):
      return "NA"
    # end if
    # sort target co-ordinates of the first block
    start1 = min(self.align1.tstart, self.align1.tend)
    end1   = max(self.align1.tstart, self.align1.tend)
    # sort target co-ordinates of the second block
    start2 = min(self.align2.tstart, self.align2.tend)
    end2   = max(self.align2.tstart, self.align2.tend)
    # sort the blocks by target (genomic) start co-ordinate
    if (start1 < start2):
      start1_s = start1
      end1_s = end1
      start2_s = start2
      end2_s = end2
    else:
      start1_s = start2
      end1_s = end2
      start2_s = start1
      end2_s = end1
    # end if
    td = int(start2_s) - int(end1_s)
    return td
  # end def

  def junction(self):
    # Determine the "junction_from" target co-ordinate
    junction_from = self.align1.tend
    # determine which side of the junction point block one represents
    if (self.align1.tstart < self.align1.tend):
      dir_from = "up"
    else:
      dir_from = "down"
    # end if

    # Determine the "junction_to" target co-ordinate
    junction_to = self.align2.tstart
    # determine which side of the junction point block one represents
    if (self.align2.tstart < self.align2.tend):
      dir_to = "down"
    else:
      dir_to = "up"
    # end if

    return "%s:%d(%s)-%s:%d(%s)" % \
      (self.align1.target, int(junction_from), dir_from,
       self.align2.target, int(junction_to),   dir_to)
  # end def

  def details(self):
    #function formats the output lines from the alignment pair
    ctg_fields = self.contig.split(" ")
    ctg = "%s(%sbp)" % (ctg_fields[0], self.contig_size)
    target = "%s:%s-%s,%s:%s-%s" % \
      (self.align1.target, self.align1.tstart, self.align1.tend,
       self.align2.target, self.align2.tstart, self.align2.tend)
    query = "%s-%s,%s-%s" % \
      (self.align1.qstart, self.align1.qend,
       self.align2.qstart, self.align2.qend)
    align_len1 = float(self.align1.qend) - float(self.align1.qstart) + 1
    align_len2 = float(self.align2.qend) - float(self.align2.qstart) + 1
    align_frac1 = align_len1/float(self.align1.query_len)
    align_frac2 = align_len2/float(self.align2.query_len)
    params = \
      "CO:%.3f,CC:%.2f,TD:%s,I1:%.1f,I2:%.1f,AF1:%.2f,AF2:%.2f" % \
      (self.overlap_query_fraction, self.contig_cover_fraction,
       self.target_distance,
       float(self.align1.identity), float(self.align2.identity),
       align_frac1, align_frac2)
    detail_string = \
      "CTG:%s TYPE:%s TARGET:%s CONTIG:%s JUNCTION:%s %s" % \
      (ctg, self.type, target, query, self.junction(), params)
    return detail_string
  # end def
# end class

def fix_align(align):
  if (hasattr(align, "fixed") and align.fixed):
    return align
  # end if
  if ("blat" == align.method):
    if ("-" == align.query_strand):
      temp_tstart = align.tstart
      align.tstart = align.tend
      align.tend = temp_tstart
    # end if
    if (None != align.psl_str):
      PSL_TNAME_COL = 13
      psl_info = align.psl_str.split("\t")
      target_name = psl_info[PSL_TNAME_COL]
      if ("chr" != target_name[0:3]):
        psl_info[PSL_TNAME_COL] = "chr%s" % target_name
        align.psl_str = "\t".join(psl_info)
      # end if
    # end if
  elif ("exonerate" == align.method):
    if (align.target_strand):
      align.query_strand = "+"
    else:
      align.query_strand = "-"
    # end if
  # end if
  align.fixed = True
  return align
# end def

#### EXCEPTION CLASSES ####
class SplitAlignmentError(Exception):
  def __init__(self, msg):
    self.msg = msg
  def __str__(self):
    return repr(self.msg)
# end class
