#!/usr/bin/perl

=head1 NAME

appendpolyAs.pl - Append poly-A and poly-T for a FASTA file.

=head1 SYNOPSYS

  % appendpolyAs.pl -i INPUT

=head1 DESCRIPTION

Reads a FASTA file and appends poly-As to the end of (+) sequences and poly-Ts
to the beginning of (-) sequences.


=head2 NOTES

This is still a work in progress.

=head1 COMMAND-LINE OPTIONS

Options can be abbreviated.  For example, you can use -s for
    --input            Input FASTA file
                         Default = STDIN
    --a                Append poly-A to all
                         Default = 1
    --t                Append poly-T to all
                         Default = 0
    --dup              Remove duplicate FASTA entries
    --length           Length of the poly-A or poly-T
                         Default = 400
    --getLength        Append length of the sequence
    --quiet            Prints out extra information on tables and results
                         Default = 0


=head1 SEE ALSO

=head1 AUTHOR

Mark Okada, hmokada@sfu.ca

Copyright (c) 2009 Simon Fraser University

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.  See DISCLAIMER.txt for
disclaimers of warranty.

=cut

use strict;
use FindBin;

use lib "$FindBin::RealBin/../lib";

use Digest::MD5 qw(md5_hex);

use Data::Dumper;
use Getopt::Long;
use Config::General;
use vars qw(%OPT %CONF);

#default vals
%OPT = (length=>400,quiet => 0);

GetOptions(\%OPT,
	   "input=s",
	   "a",
	   "t",
	   "dup",
	   "length=i",
	   "getLength",
	   "quiet+");


open (FH, $OPT{input}) if $OPT{input};
my $fh = -e $OPT{input} ? *FH : *STDIN;


my $header;
my @seq;

while (<$fh>) {
    chomp;
    
    if (substr($_,0,1) eq ">") {
        proc_entry($header,\@seq) if $header;
        @seq = ();
        $header = $_;
    } else {
        push @seq, $_;
    }
    
    
}

if ($header) {
    proc_entry($header, \@seq);
}


close FH;




sub proc_entry {
    my $header = shift;
    my $seqs = shift;
    
    
    # check for duplicates
    if ($OPT{dup}) {
	my $digest = md5_hex(join("",@$seqs));
	
	$OPT{digests}{$digest}++;
	return if 1 < $OPT{digests}{$digest};
    }
    
    
    # append the length to the first word of the header
    if ($OPT{getLength}) {
	my ($elem,$other) = $header =~ /^(\S+)(.*)/;
	$header = $elem
	    ."__length=" 
	    . (length(join("",@$seqs)) + ($OPT{length} * ($OPT{a} + $OPT{t}) ))
	    . $other;;
    }
    
    
    print "$header\n";
    #print "@$seqs\n\n";
    
    if ($OPT{a}) {
	push @$seqs,"A" x $OPT{length};
    }
    
    if ($OPT{t}) {
	unshift @$seqs, "T" x $OPT{length};
    }
    
    printfasta(join("",@$seqs));
    
}


sub printfasta {
    my $seq = shift;

    # put into FA format, break sequence up into lines
    my @out = unpack "a80" x _ceil(length($seq) / 80), $seq;

    for (@out) {print $_, "\n"}
    
}


sub _ceil {
    my $v = shift;
    my $f = int($v);

    return $f if $v == $f || $v < 0;
    return int($v+1);
}



1;
