#!/usr/bin/perl

=head1 NAME

findMatePairedMappings.pl - combine a list of mapped TSV files for each mate pair

=head1 SYNOPSYS

  % findMatePairedMappings.pl -r READFILE -m MATEFILE
=head1 DESCRIPTION

Combine two lists of mapped reads in a a tab delimited file given from samse2csv.pl program.
The input files can be gzipped.

=head2 NOTES

This script uses backticks to extract the information from the files.
This is still a work in progress.

=head1 COMMAND-LINE OPTIONS

Options can be abbreviated.  For example, you can use -r for read
    --read             Input read file
    --mate             Input mate read file
    --trim             Consider only trimmed reads, read names start with "@tr"
    --dmin             Minimum distance allowed for a mate pair mapping
                         Default = 100
    --dmax             Maximum distance allowed for a mate pair mapping
                         Default = 250
    --all              Look at all possible transcripts mapped instead of best 1
    --quiet            Prints out extra information on tables and results
                         Default = 0


=head1 SEE ALSO

=head1 AUTHOR

Mark Okada, hmokada@sfu.ca

Copyright (c) 2010 Simon Fraser University

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.  See DISCLAIMER.txt for
disclaimers of warranty.

=cut

use strict;
use FindBin;

use lib "$FindBin::RealBin/../lib";

use Data::Dumper;
use Getopt::Long;
use Config::General;
use vars qw(%OPT %CONF);

#default vals
%OPT = (dmin => 100,
	dmax => 250,
	quiet => 0);

GetOptions(\%OPT,
	   "read=s",
	   "mate=s",
	   "trim",
	   "dmin=i",
	   "dmax=i",
	   "all",
	   "quiet+");


# [trimmed] reads: readname -> @(transcript, offset, strand)
my $reads;

my $cmd = guess_extension($OPT{read}) . " $OPT{read}";

for my $line (`$cmd`) {
    chomp $line;
    my @tok = split "\t", $line;
    my ($prefix) = $tok[0] =~ /^(\S+)/;
    
    
    if (substr($prefix,1,2) eq "tr") {
	$prefix = ">" . substr($prefix,3);
    } elsif ($OPT{trim}) {
	next;
    }
    
    
    push @{$reads->{$prefix}}, [@tok[1..8]];

}



# go through the ET's mate read file
my @mateset;
my $curr;
for my $line (`less $OPT{mate}`) {
    chomp $line;
    my @tok = split "\t", $line;
    ($tok[0]) = $tok[0] =~ /^(\S+)/;
    
    if ($tok[0] eq $curr) {
	
	push @mateset, \@tok;
    } else {
	
	compare_two_sets($curr, \@mateset, $reads) if @mateset && $reads->{$curr};
	
	@mateset = (\@tok);
	$curr = $tok[0];
    }
    
}
compare_two_sets($curr, \@mateset, $reads) if @mateset && $reads->{$curr};


sub compare_two_sets {
    my $prefix = shift;
    my $mateset = shift;
    my $reads = shift;
    
    my @output;
    
  OUT:
    for my $readentry (@{$reads->{$prefix}}) {
	for my $mateentry (@{$mateset}) {
	    
	    my $absdist = abs($readentry->[4] - $mateentry->[5]);
	    
	    # legitimate read: opp strand, same transcript, maps between 100-250bp
	    if ($readentry->[0] eq $mateentry->[1] &&
		$readentry->[6] ne $mateentry->[7] &&
		$absdist < $OPT{dmax} &&
		$absdist > $OPT{dmin})
	    {
		push @output, join("\t",$prefix,@{$readentry}),"\n";
		
		if ($OPT{all}) {
		    last;
		} else {
		    last OUT;
		}
		
	    }
	}
    }
    
    print @output;
    
}


sub guess_extension {
    my $filename = shift;
    my $ext = substr($filename, -3);
    if ($ext eq ".gz") {return "zcat";}
    elsif ($ext eq "bz2") {return "bzcat";}
    else {return "cat";}
}

    

1;
