#!/usr/bin/perl

=head1 NAME

polyafinder.pl - A set of scripts to extract polyadenylation sites for transcripts

=head1 SYNOPSYS

  % polyafinder.pl [-c CONFIGFILE]

=head1 DESCRIPTION

A wrapper for a set of scripts which finds the polyadenylation sites (known and novel)
of transcripts using paired end whole transcriptome shotgun sequencing (WTSS) reads.


=head2 NOTES

This is still a work in progress.

=head1 COMMAND-LINE OPTIONS

Options can be abbreviated.  For example, you can use -i for input
    --forward          Forward reads containing PAM or EJ reads
    --reverse          Reverse reads containing PAM or EJ reads
    --mf               Mate read of the reverse EJ read (forward dir)
    --mr               Mate read of the forward EJ read (rev dir)
    --clean            Reconstruct all transcript & contig BWA files
                         Default = 0 (does not index if BWT file present)
    --transcript       Transcript file to be mapped
    --append           Append poly-A's to the transcript
    --stem             Output filename stem
                         Default = "polya_out";
    --contig           Contig file to be mapped (PAM only)
    --cappend          Append poly-A's to contigs as well as poly-T's to start
    --configfile       Configuration file which includes all parameters
                         Default = "../conf/polyafinder.conf"
    --ej               Boolean to process end junction (EJ)  reads
                         Default = 0 (PAM read mode on)
    --dev              Prints out commands only, does not run anything
    --quiet            Prints out extra information on tables and results
                         Default = 0


=head1 SEE ALSO

polyareads.pl polyaanalyse.pl, BWA

=head1 AUTHOR

Mark Okada, hmokada@sfu.ca

Copyright (c) 2010 Simon Fraser University

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.  See DISCLAIMER.txt for
disclaimers of warranty.

=cut

use strict;
use FindBin;

use lib "$FindBin::RealBin/../lib";

use Data::Dumper;
use Getopt::Long;
use Config::General;
use vars qw(%OPT %CONF);

#default vals
%OPT = (stem => "polya_out",
	quiet => 0);

my @transcript_files;


GetOptions(\%OPT,
	   "forward=s",
	   "reverse=s",
	   "mf=s",
	   "mr=s",
	   "clean",
	   "transcript=s"=>\@transcript_files,
	   "append",
	   "stem=s",
	   "contig=s",
	   "cappend",
	   "congifile=s",
	   "ej",
	   "dev",
	   "quiet+");

%CONF = Config::General::ParseConfig(-ConfigFile=>$OPT{configfile}
                                     || "$FindBin::RealBin/../conf/polyafinder.conf") or die $!;


my @cmds;
my @comments;








### Files - intermediate and final output


# transcript files
my $transcript;
# requires polyA tails?
if ($OPT{append}) {
    $transcript = $OPT{stem} . ".transPolyA.fa";
    for (@transcript_files) {
	die "Input transcript file $_ same as poly-A appended file, turn off -append option" if $_ eq $transcript;
    }
    
    # append polyA tails to the transcript - assumed that transcripts are sense
    push @comments, "Appending poly-A file $transcript from transcript file(s) @transcript_files";
    push @cmds, join(" ",
		     guess_extension(@transcript_files),
		     @transcript_files,
		     "|",
		     "$CONF{dir}{bin}/appendpolyAs.pl",
		     "-a -length $CONF{read_length} -getLength -dup",
		     ">",
		     $transcript);
} else {
    die "Must provide exactly one transcript if polyA's not added (-a switch)" unless @transcript_files == 1;
    ($transcript) = @transcript_files;
    if (guess_extension($transcript) ne "cat") {
	my $new_transfile = $OPT{stem} . ".transPolyA.fa";
	push @comments, "Unzipping transcript file to $new_transfile";
	push @cmds, guess_extension($transcript) . " $transcript > $new_transfile";
	$transcript = $new_transfile;
    }
}

# index transcript file if required
if (! $OPT{clean} && -e "$transcript.bwt") {
    push @comments, "Transcript file $transcript already indexed (BWA)";
    push @cmds, "";
} else {
    push @comments, "Indexing transcript file $transcript";
    push @cmds, "$CONF{dir}{bwa}/bwa index -a bwtsw $transcript";
}


run_batch(\@comments, \@cmds);

# check success
die "Error in transcript file" if !$OPT{dev} && (! -r $transcript || -z $transcript);





# Poly-A mate (PAM) reads
if (! $OPT{ej}) {
    my $forward_aln  = $OPT{stem} . ".pam.aln.forward.sai";
    my $reverse_aln = $OPT{stem} . ".pam.aln.reverse.sai";
    
    my $forward_unmapped  = $OPT{stem} . ".pam.aln.forward.unmapped";
    my $reverse_unmapped = $OPT{stem} . ".pam.aln.reverse.unmapped";
    
    my $forward_tsv  = $OPT{stem} . ".pam.aln.forward.tsv";
    my $reverse_tsv = $OPT{stem} . ".pam.aln.reverse.tsv";
    
    
    push @comments, "Aligning forward reads to transcripts";
    push @cmds, join(" ",
		     "$CONF{dir}{bwa}/bwa",
		     "aln",
		     $transcript,
		     $OPT{forward},
		     ">",
		     $forward_aln);
    
    push @comments, "Aligning reverse reads to transcripts";
    push @cmds, join(" ",
		     "$CONF{dir}{bwa}/bwa",
		     "aln",
		     $transcript,
		     $OPT{reverse},
		     ">",
		     $reverse_aln);
    
    push @comments, "Saving forward read alignments to TSV format";
    push @cmds, join(" ",
		     "$CONF{dir}{bwa}/bwa",
		     "samse",
		     "-n $CONF{bwa}{max_results}",
		     $transcript,
		     $forward_aln,
		     $OPT{forward},
		     "|",
		     "$CONF{dir}{bin}/samse2csv.pl",
		     "-fromend",
		     "-u $forward_unmapped",
		     ">",
		     $forward_tsv);
    
    push @comments, "Saving reverse read alignments to TSV format";
    push @cmds, join(" ",
		     "$CONF{dir}{bwa}/bwa",
		     "samse",
		     "-n $CONF{bwa}{max_results}",
		     $transcript,
		     $reverse_aln,
		     $OPT{reverse},
		     "|",
		     "$CONF{dir}{bin}/samse2csv.pl",
		     "-fromend",
		     "-u $reverse_unmapped",
		     ">",
		     $reverse_tsv);
    
    run_batch(\@comments,\@cmds);
    
    # output the filenames if successful
    if ($OPT{dev} || (-e $forward_tsv && $reverse_tsv)) {
	print join ("\n",
		    "################################",
		    "# Created PAM alignment table output:",
		    "# Forward PAM alignments: $forward_tsv",
		    "# Reverse PAM alignments: $reverse_tsv",
		    "#"),"\n\n";
    }
    
    # check success
    warn "Error in forward alignment file" if (!$OPT{dev} && (! -r $forward_tsv || -z $forward_tsv));
    warn "Error in reverse alignment file" if (!$OPT{dev} && (! -r $reverse_tsv || -z $reverse_tsv));
    
    
# align with contigs - look for lengthenings
    my $contig;
    
    if ($OPT{contig}) {
	if ($OPT{cappend}) {
	    $contig = $OPT{stem}. ".contigPolyAT.fa";
	    push @comments, "Appending polyA and polyT to contigs";
	    push @cmds, join(" ",
			     guess_extension($OPT{contig}),
			     $OPT{contig},
			     "|",
			     "$CONF{dir}{bin}/appendpolyAs.pl",
			     "-t -a",
			     "-length $CONF{read_length}",
			     "-getLength",
			     ">",
			     $contig);
	} else {
	    die "Contig file $OPT{contig} not found" if ! -r $OPT{contig};
	    $contig = $OPT{contig};
	    if (guess_extension($contig) ne "cat") {
		my $new_contigfile = $OPT{stem} . ".contigPolyAT.fa";
		push @comments, "Unzipping contig file to $new_contigfile";
		push @cmds, guess_extension($contig) . " $contig > $new_contigfile";
		$contig = $new_contigfile;
	    }
	}
	
	
	# index contig file is required
	if (! $OPT{clean} && -e "$contig.bwt") {
	    push @comments, "Contig file $contig already indexed (BWA)";
	    push @cmds, "";
	} else {
	    push @comments, "Indexing contig file $contig";
	    push @cmds, "$CONF{dir}{bwa}/bwa index -a bwtsw $contig";
	}
	
	
	run_batch(\@comments, \@cmds);
	
	
	# align to contigs
	my $forward_c_aln  = $OPT{stem} . ".pam.contig.aln.forward.sai";
	my $reverse_c_aln = $OPT{stem} . ".pam.contig.aln.reverse.sai";
	
	my $forward_c_unmapped  = $OPT{stem} . ".pam.contig.aln.forward.unmapped";
	my $reverse_c_unmapped = $OPT{stem} . ".pam.contig.aln.reverse.unmapped";
	
	my $forward_c_tsv  = $OPT{stem} . ".pam.contig.aln.forward.tsv";
	my $reverse_c_tsv = $OPT{stem} . ".pam.contig.aln.reverse.tsv";
	
	my $forward_c_remapped  = $OPT{stem} . ".pam.contig.remapped.forward.tsv";
	my $reverse_c_remapped = $OPT{stem} . ".pam.contig.remapped.reverse.tsv";
	
	push @comments, "Aligning forward reads to contigs";
	push @cmds, join(" ",
			 "$CONF{dir}{bwa}/bwa",
			 "aln",
			 $contig,
			 $OPT{forward},
			 ">",
			 $forward_c_aln);
	
	push @comments, "Aligning reverse reads to contigs";
	push @cmds, join(" ",
			 "$CONF{dir}{bwa}/bwa",
			 "aln",
			 $contig,
			 $OPT{reverse},
			 ">",
			 $reverse_c_aln);
	
	push @comments, "Saving forward read alignments to TSV format";
	push @cmds, join(" ",
			 "$CONF{dir}{bwa}/bwa",
			 "samse",
			 "-n $CONF{bwa}{max_results}",
			 $contig,
			 $forward_c_aln,
			 $OPT{forward},
			 "|",
			 "$CONF{dir}{bin}/samse2csv.pl",
			 "-fromend",
			 "-u $forward_c_unmapped",
			 ">",
			 $forward_c_tsv);
	
	push @comments, "Saving reverset read alignments to TSV format";
	push @cmds, join(" ",
			 "$CONF{dir}{bwa}/bwa",
			 "samse",
			 "-n $CONF{bwa}{max_results}",
			 $contig,
			 $reverse_c_aln,
			 $OPT{reverse},
			 "|",
			 "$CONF{dir}{bin}/samse2csv.pl",
			 "-fromend",
			 "-u $reverse_c_unmapped",
			 ">",
			 $reverse_c_tsv);
	
	
	# find those mapped reads which were unmapped previously for transcripts
	push @comments, "Find forward reads that map only to contigs";
	push @cmds, join(" ",
			 "$CONF{dir}{bin}/getallunmappedfromTSV.pl",
			 $forward_unmapped,
			 $forward_c_tsv,
			 ">",
			 $forward_c_remapped);
	
	push @comments, "Find reverset reads that map only to contigs";
	push @cmds, join(" ",
			 "$CONF{dir}{bin}/getallunmappedfromTSV.pl",
			 $reverse_unmapped,
			 $reverse_c_tsv,
			 ">",
			 $reverse_c_remapped);
	
	run_batch(\@comments, \@cmds);
	
	# output the filenames if successful
	if ($OPT{dev} || (-e $forward_c_remapped && -e $reverse_c_remapped)) {
	    print join ("\n",
			"################################",
			"# Created PAM contig alignment table output:",
			"# Forward PAM contig alignments: $forward_c_remapped",
			"# Reverse PAM contig alignments: $reverse_c_remapped",
			"#"),"\n\n";
	}
	
    }
}


# End junction reads
else {
    # calculate the number of poly-T's on each EJ read
    
    my $ej_tailleng_file = $transcript . ".taillengths.txt";
    
    my $forward_aln  = $OPT{stem} . ".ej.aln.forward.sai";
    my $reverse_aln = $OPT{stem} . ".ej.aln.reverse.sai";
    
    my $forward_tsv  = $OPT{stem} . ".ej.aln.forward.tsv";
    my $reverse_tsv = $OPT{stem} . ".ej.aln.reverse.tsv";
    
    my $forward_stats = $OPT{stem} . ".ej.aln.stats.forward.txt";
    my $reverse_stats = $OPT{stem} . ".ej.aln.stats.reverse.txt";
    
    push @comments, "Calculating number of poly-A's in each transcript";
    push @cmds, join(" ",
		     "$CONF{dir}{bin}/calcPolyAinTails.pl",
		     "-i $transcript",
		     "-o -$CONF{read_length}",
		     ">",
		     $ej_tailleng_file);
    
    
    push @comments, "Aligning forward EJ read to transcript";
    push @cmds, join(" ",
		     "$CONF{dir}{bwa}/bwa",
		     "aln",
		     $transcript,
		     $OPT{forward},
		     ">",
		     $forward_aln);
    
    push @comments, "Aligning reverse EJ read to transcript";
    push @cmds, join(" ",
		     "$CONF{dir}{bwa}/bwa",
		     "aln",
		     $transcript,
		     $OPT{reverse},
		     ">",
		     $reverse_aln);
    
    push @comments, "Extracting forward reads that have only trimmed reads align";
    push @cmds, join(" ",
		     "$CONF{dir}{bwa}/bwa",
		     "samse",
		     "-n $CONF{bwa}{max_results}",
		     $transcript,
		     $forward_aln,
		     $OPT{forward},
		     "|",
		     "$CONF{dir}{bin}/samse2csv.pl",
		     "-fromend",
		     "-offset",
		     $ej_tailleng_file,
		     "-endread",
		     "-negalign",
		     "|",
		     "$CONF{dir}{bin}/procendtagtsv.pl",
		     "-s $forward_stats",
		     "-a",
		     ">",
		     $forward_tsv);
    
    push @comments, "Extracting reverse reads that have only trimmed reads align";
    push @cmds, join(" ",
                     "$CONF{dir}{bwa}/bwa",
                     "samse",
                     "-n $CONF{bwa}{max_results}",
                     $transcript,
                     $reverse_aln,
                     $OPT{reverse},
                     "|",
                     "$CONF{dir}{bin}/samse2csv.pl",
                     "-fromend",
                     "-offset",
                     $ej_tailleng_file,
                     "-endread",
                     "-negalign",
                     "|",
                     "$CONF{dir}{bin}/procendtagtsv.pl",
                     "-s $reverse_stats",
                     "-a",
                     ">",
                     $reverse_tsv);
    
    
    
    run_batch(\@comments,\@cmds);
    
    
    # output the filenames if successful
    if ($OPT{dev} || (-e $forward_tsv && -e $reverse_tsv)) {
	print join ("\n",
		    "################################",
		    "# Created EJ alignment table output:",
		    "# Forward EJ alignments: $forward_tsv",
		    "# Reverse EJ alignments: $reverse_tsv",
		    "#"),"\n\n";
    }
    
    
    
    # check success
    warn "Error in forward alignment file" if (!$OPT{dev} && (! -r $forward_tsv || -z $forward_tsv));
    warn "Error in reverse alignment file" if (!$OPT{dev} && (! -r $reverse_tsv || -z $reverse_tsv));
    
    
    # additionally, filter for EJ reads in which their mate maps to same transcript
    if ($OPT{mf} || $OPT{mr}) {
	
	my $forward_m_aln = $OPT{stem} . ".ej.aln.mate.forward.sai";
	my $reverse_m_aln = $OPT{stem} . ".ej.aln.mate.reverse.sai";
	
	my $forward_m_tsv  = $OPT{stem} . ".ej.aln.mate.forward.tsv";
	my $reverse_m_tsv = $OPT{stem} . ".ej.aln.mate.reverse.tsv";
	
	my $forward_m_stats = $OPT{stem} . ".ej.aln..stats.mate.forward.txt";
	my $reverse_m_stats = $OPT{stem} . ".ej.aln.stats.mate.reverse.txt";
	
	my $forward_match_tsv = $OPT{stem} . ".ej.aln.matematched.forward.tsv";
	my $reverse_match_tsv =$OPT{stem} . ".ej.aln.matematched.reverse.tsv";
	
	
	push @comments, "Aligning forward reads which are mates of reverse EJ reads";
	push @cmds, join(" ",
			 "$CONF{dir}{bwa}/bwa",
			 "aln",
			 $transcript,
			 $OPT{mf},
			 ">",
			 $forward_m_aln);
	
	
	push @comments, "Aligning reverse reads which are mates of forward EJ reads";
	push @cmds, join(" ",
                         "$CONF{dir}{bwa}/bwa",
                         "aln",
                         $transcript,
                         $OPT{mr},
                         ">",
                         $reverse_m_aln);
	
	push @comments, "Saving forward mate reads to TSV file";
	push @cmds, join(" ",
			  "$CONF{dir}{bwa}/bwa",
			  "samse",
			  "-n $CONF{bwa}{max_results}",
			  $transcript,
			  $forward_m_aln,
			  $OPT{mf},
			  "|",
			  "$CONF{dir}{bin}/samse2csv.pl",
			  "-fromend",
			  "-offset $ej_tailleng_file",
			  ">",
			  $forward_m_tsv);
	
	push @comments, "Saving reverse mate reads to TSV file";
        push @cmds, join(" ",
                          "$CONF{dir}{bwa}/bwa",
                          "samse",
                          "-n $CONF{bwa}{max_results}",
                          $transcript,
                          $reverse_m_aln,
                          $OPT{mr},
                          "|",
                          "$CONF{dir}{bin}/samse2csv.pl",
                          "-fromend",
                          "-offset $ej_tailleng_file",
                          ">",
                          $reverse_m_tsv);
	
	
	push @comments, "Find forward matching mate reads and save to TSV";
	push @cmds, join(" ",
			 "$CONF{dir}{bin}/findMatePairedMappings.pl",
			 "-r $reverse_tsv",
			 "-m $forward_m_tsv",
			 "-a",
			 "-dmin $CONF{mate_min}",
			 "-dmax $CONF{mate_max}",
			 ">",
			 $forward_match_tsv);
	
	
	push @comments, "Find reverse matching mate reads and save to TSV";
        push @cmds, join(" ",
			 "$CONF{dir}{bin}/findMatePairedMappings.pl",
			 "-r $forward_tsv",
			 "-m $reverse_m_tsv",
			 "-a",
			 "-dmin $CONF{mate_min}",
                         "-dmax $CONF{mate_max}",
			 ">",
			 $reverse_match_tsv);
    
	
	
	run_batch(\@comments,\@cmds);
	
	# output the filenames if successful
	if ($OPT{dev} || (-e $forward_match_tsv && -e $reverse_match_tsv)) {
	    print join ("\n",
			"################################",
			"# Filtered EJ alignments which have corresponding mate reads:",
			"# Forward EJ alignments: $forward_match_tsv",
			"# Reverse EJ alignments: $reverse_match_tsv",
			"#"),"\n\n";
	}
    }
    
}




sub run_batch {
    my $comments = shift;
    my $cmds = shift;
    
    # run commands, and print comments
    for (my $i=0; $i<@$cmds; $i++) {
	print "# ", $comments->[$i], "\n" unless $OPT{quiet};
	print $cmds->[$i], "\n" unless $OPT{quiet};
	`$cmds->[$i]` unless $OPT{dev};
    }
    
    @$comments = ();
    @$cmds = ();
    
}


sub guess_extension {
    my $filename = shift;
    my $ext = substr($filename, -3);
    if ($ext eq ".gz") {return "zcat";}
    elsif ($ext eq "bz2") {return "bzcat";}
    else {return "cat";}
}


1;
