#!/usr/bin/perl

=head1 NAME

polyareads.pl - Extract polyadenylation junction reads from solexa files

=head1 SYNOPSYS

    % polyareads.pl -l FORWARDREADS1 [-l FORWARDREADS2...] -r REVREADS1 [-r REVREADS2...] [-c CONFIGFILE]

=head1 DESCRIPTION

A wrapper for a set of scripts which finds the polyadenylation sites (known and novel)
of transcripts using paired end whole transcriptome shotgun sequencing (WTSS) reads.

Steps

=over

=item -Condense all raw reads as FASTQ for forward and reverse reads ("1" and "2")

=item -Look at each pair, extract all those with PAM reads with given cutoff

=head2 NOTES

The script will attempt to use "cat", "zcat", and "bzcat" depending on the file extension.

The parameters are controlled by changing the configuration file.

"forward" read referrs to read 1, "reverse" referrs to read 2.

This is still a work in progress.

=head1 COMMAND-LINE OPTIONS

Options can be abbreviated.  For example, you can use -i for input
    --forward          forward reads
    --reverse          reverse reads
    --stem             Output filename stem
                         Default = "polya_out"
    --configfile       Configuration file which includes all parameters
                         Default = "../conf/polyareads.conf"
    --ej               Boolean to process end junction (EJ)  reads
    --pam              Boolean to process poly-A mate (PAM) reads
    --dev              Prints out commands only, does not run anything
    --quiet            Prints out extra information on tables and results
                         Default = 0


=head1 SEE ALSO

polyafinder.pl

=head1 AUTHOR

Mark Okada, hmokada@sfu.ca

Copyright (c) 2010 Simon Fraser University

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.  See DISCLAIMER.txt for
disclaimers of warranty.

=cut

use strict;
use FindBin;

use lib "$FindBin::RealBin/../lib";

use Data::Dumper;
use Getopt::Long;
use Config::General;
use File::Temp qw(tempfile);

use vars qw(%OPT %CONF);

#default vals
%OPT = (output => "polya_out",
	quiet => 0);


# input forward and reverse files
my @forward_files;
my @reverse_files;

# final output files genereated
my @outputfiles;


GetOptions(\%OPT,
	   "forward=s"=>\@forward_files,
	   "reverse=s"=>\@reverse_files,
	   "output=s",
	   "configile=s",
	   "ej",
	   "pam",
	   "dev",
	   "quiet+");

%CONF = Config::General::ParseConfig(-ConfigFile=>$OPT{configfile}
				     || "$FindBin::RealBin/../conf/polyareads.conf") or die $!; 

if (! $OPT{ej} && ! $OPT{pam}) {
    die "Please select either PAM or EJ reads -e or -p";
}


my @cmds;
my @comments;


# other file checks
die "There must be matching forward and reverse read files" if @forward_files != @reverse_files;
for (@forward_files, @reverse_files) {
    die "Cannot read file $_" if ! -r $_;
}



if ($OPT{pam}) {

    my $forward_output     = "$OPT{output}.pam.1.forwardT.fastq";
    my $reverse_output     = "$OPT{output}.pam.2.reverseT.fastq";
    
    if (-e $forward_output || -e $reverse_output) {
	push @comments, "Delete output FASTQ files";
	push @cmds,"rm $forward_output $reverse_output";
    }
    
    
    for (my $i=0; $i<@forward_files; $i++) {
	my $forward_file  = $forward_files[$i];
	my $reverse_file = $reverse_files[$i];
	
	push @comments, "Extracting forward reads $forward_file";
	push @cmds, join(" ",
			 "$CONF{dir}{bin}/extractCutoffPolyA.pl",
			 "-a",
			 "-c", $CONF{pam_cutoff},
			 "-f", $forward_file,
			 "-r", $reverse_file,
			 "-of", $forward_output,
			 "-or", $reverse_output);
	
    }
    
    # output the filenames if successful
    if ($OPT{dev} || (-e $forward_output && $reverse_output)) {
	push @outputfiles, join ("\n",
				 "################################",
				 "# Created PAM read outputs:",
				 "# Forward PAM reads: $forward_output",
				 "# Reverse PAM reads: $reverse_output",
				 "#",
				 "# To do alignment, use command:",
				 "# polyafinder.pl -f $forward_output -r $reverse_output -t <TRANSCRIPT> -a [ -cont <CONTIGFILE> -capp]\n\n");
    
    }
    
}



# End junction reads                                                                                                   
if ($OPT{ej}) {
    my $forward_output     = "$OPT{output}.ej.1.forwardT.fastq";
    my $reverse_output     = "$OPT{output}.ej.2.reverseT.fastq";
    my $forward_outputmate = "$OPT{output}.ej.1.mate.fastq";
    my $reverse_outputmate = "$OPT{output}.ej.2.mate.fastq";
    
    if (-e $forward_output || -e $reverse_output || -e $forward_outputmate || -e $reverse_outputmate) {
        push @comments, "Delete output FASTQ files";
	push @cmds,"rm $forward_output $reverse_output $forward_outputmate $reverse_outputmate";
    }
    
    
    for (my $i=0; $i<@forward_files; $i++) {
        my $forward_file  = $forward_files[$i];
        my $reverse_file = $reverse_files[$i];
	
        push @comments, "Extracting forward reads $forward_file";
        push @cmds, join(" ",
                         guess_extension($forward_file),
                         $forward_file,
                         "|",
                         "$CONF{dir}{bin}/getrawreadendtag.pl -n",
                         ">>",
                         $forward_output);

        push @comments, "Extracting reverse EJ reads $reverse_file";
        push @cmds, join(" ",
                         guess_extension($reverse_file),
                         $reverse_file,
                         "|",
                         "$CONF{dir}{bin}/getrawreadendtag.pl -n",
                         ">>",
                         $reverse_output);
    }
    
    
    my $ej_readnames = "$OPT{output}.ej.readnames.tsv";
    
    push @comments, "Getting mate read names forward-T";
    push @cmds, join(" ",
		     "cat",
		     $forward_output,
		     $reverse_output,
		     "|",
		     "$CONF{dir}{bin}/getejreadnames.pl",
		     ">",
		     $ej_readnames);
    
    for (my $i=0; $i<@forward_files; $i++) {
        my $forward_file  = $forward_files[$i];
        my $reverse_file = $reverse_files[$i];
	
	push @comments, "Extracting mate of EJ read pairs forward-T";
	push @cmds, join(" ",
			 guess_extension($forward_file),
			 $forward_file,
			 "|",
			 "$CONF{dir}{bin}/extractmatereads.pl",
			 $ej_readnames,
			 ">>",
			 $forward_outputmate);
	
	push @comments, "Extracting mate of EJ reads reverse-T";
	push @cmds, join(" ",
			 guess_extension($reverse_file),
			 $reverse_file,
			 "|",
			 "$CONF{dir}{bin}/extractmatereads.pl",
			 $ej_readnames,
			 ">>",
			 $reverse_outputmate);
    }
    
    
    if ($OPT{dev} || (-e $forward_output     && -e $reverse_output    &&
		      -e $forward_outputmate && -e $reverse_outputmate)) {
	push @outputfiles, join ("\n",
				 "################################",
				 "# Created EJ read outputs:",
				 "# Forward EJ read: $forward_output",
				 "# Reverse EJ read: $reverse_output",
				 "# Forward EJ read-mate: $forward_outputmate",
				 "# Reverse EJ read-mate: $reverse_outputmate",
				 "#",
				 "# To do alignment, use command:",
				 "# polyafinder.pl -e -f $forward_output -r $reverse_output -mf $forward_outputmate -mr $reverse_outputmate -t <TRANSCRIPT>\n\n");
    }
}





# run commands, and print comments
for (my $i=0; $i<@cmds; $i++) {
    print "# ", $comments[$i], "\n" unless $OPT{quiet};
    print $cmds[$i], "\n" unless $OPT{quiet};
    `$cmds[$i]` unless $OPT{dev};
}



print @outputfiles unless $OPT{quiet};




# guess the program used to open a file: txt, gz, bz2
sub guess_extension {
    my $filename = shift;
    my $ext = substr($filename, -3);
    if ($ext eq ".gz") {return "zcat";}
    elsif ($ext eq "bz2") {return "bzcat";}
    else {return "cat";}
}


1;
