#!/usr/bin/perl

=head1 NAME

procendtagtsv.pl - Reads a TSV file for End reads, only consider top hits for trimmed reads

=head1 SYNOPSYS

  % cat TSVFILE | procendtagtsv.pl > OUTPUTTSV

=head1 DESCRIPTION

Reads a TSV file for end read specific cases. This does two things, keeps only the top hit for
each alignment, and it removes cases where both the raw and trimmed read map to the same location,
meaning that the read is not evidence for a novel poly adenylation site.

=head2 NOTES

This is still a work in progress.

=head1 COMMAND-LINE OPTIONS

Options can be abbreviated.  For example, you can use -s for
    --offset           Adjust the alignment by the number of polyA in actual seq - input sep file
    --t                Uses the 3rd column of the offset file if read is mapped to antisense
    --contig           Contig mode, ignore trimmed "tr" contigs
    --stats            Output stats for the end tags
    --readlength       Readlength of the reads, cutoff of how close raw and trimmed reads can get
                         Default = 50
    --all              Exhaustively look for all mappings for a read instead of just the best
    --quiet            Prints out extra information on tables and results
                         Default = 0


=head1 SEE ALSO

=head1 AUTHOR

Mark Okada, hmokada@sfu.ca

Copyright (c) 2009 Simon Fraser University

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.  See DISCLAIMER.txt for
disclaimers of warranty.

=cut

use strict;
use FindBin;

use lib "$FindBin::RealBin/../lib";

use Data::Dumper;
use Getopt::Long;
use Config::General;
use vars qw(%OPT %CONF);

#default vals
%OPT = (readlength=>50, quiet => 0);

GetOptions(\%OPT,
	   "stats=s",
	   "offset=s",
	   "t",
	   "contig",
	   "readlength=i",
	   "all",
	   "quiet+");

my $stats = {num_both_map => 0};

my @raw;
my @trim;
my $prev = "";
my $offset;

# adjust the tail length by the offset file
open (OFF, $OPT{offset}) if $OPT{offset};
while (<OFF>) {
    chomp;
    my ($read,@vals) = split "\t", substr($_,1);
    $read =~ s/__.+//;
    $read =~ s/ .+//;
    $offset->{$read} = [@vals];
}
close OFF;


open (ST, ">$OPT{stats}") if $OPT{stats};


while (<>) {
    chomp;
    my @tok = split "\t";
    
    if (substr($tok[0],0,3) eq ">tr") {
	# process trimmed alignment
	($tok[0]) = $tok[0] =~ />tr(\S+)/;
	$tok[0] = ">" . $tok[0];
	
	if ($tok[0] eq $prev) {
	    push @trim, \@tok;
	} else {
	    # diff read name, start a new buffer set
	    proc_readsets(\@raw, \@trim, $stats, $offset);
	    @raw  = ();
	    @trim = ();
	    
	    push @trim, \@tok;
	    $prev = $tok[0];
	}
	
    } else {
	# process raw untrimmed alignment
	($tok[0]) = $tok[0] =~ /^(\S+)/;
	
	if ($tok[0] eq $prev) {
	    push @raw, \@tok;
	} else {
	    # diff read name, start a new buffer set
	    proc_readsets(\@raw, \@trim, $stats, $offset);
	    @raw  = ();
	    @trim = ();
	    
	    push @raw, \@tok;
	    $prev = $tok[0];
	}
	
    }
    
    
    
}
if ($OPT{stats}) {
    for my $key (sort keys %$stats) {
	print ST $key, "\t", $stats->{$key}, "\n";
    }
    close ST;
}

sub proc_readsets {
    if ($OPT{contig}) {
	proc_readsets_contigs(@_);
    } else {
	proc_readsets_paired(@_);
    }
}

sub proc_readsets_paired {
    my $raw = shift;
    my $trim = shift;
    my $stats = shift;
    my $offset = shift;
    
    return if !@$trim;
    $stats->{num_pairs}++;
    
    my @output;
    my $ok = 1;
    
    # check to see if raw read aligns to same place as trimmed read -> fp
    while (my $ttok = shift @$trim) {
	
	# only use first entry if not selected
	@$trim = () unless ($OPT{all});
	
	#my $ttok = shift @$trim;
	my $fail = 0;
	for my $rtok (@$raw) {
	    
	    
	    if ($$ttok[1] eq $$rtok[1]  && abs($$ttok[5] - $$rtok[5]) < $OPT{readlength}) {
		$fail++;
	    }
	}
	
	
	# check if there are any raw reads
	if (scalar @$raw > 0) {
	    $stats->{num_raw_maps}++;
	    
	}
	if ($fail) {
	    $stats->{num_both_map}++;
	    $ok = 0;
	} else {
	    $$ttok[5] -= $offset->{$$ttok[1]}->[0] if $OPT{offset};
	    push @output, join("\t", @$ttok), "\n";
	}
	
    }
    
    print @output if $ok;
    
}

sub proc_readsets_contigs {
    my $raw = shift;
    my $trim = shift;
    my $stats = shift;
    my $offset = shift;
    
    if (@$raw) {
	$stats->{num_raw}++;
    }
    
    
    my @output;
    
    while (my $ttok = shift @$trim) {
	
        # only use first entry if not selected
        @$trim = () unless ($OPT{all});
	
	if ($OPT{offset}) {
	    if (! $OPT{t}
		|| (0 < $$ttok[8] && $$ttok[7] eq "+")
		|| ($$ttok[8] < 0 && $$ttok[7] eq "-") ) {
		$$ttok[5] -= $offset->{$$ttok[1]}->[0];
	    } else {
		$$ttok[5] -= $offset->{$$ttok[1]}->[1];
	    }
	}
	push @output, join("\t", @$ttok), "\n";
	
	
	
    }
    
    print @output;
    
}



1;
