#!/usr/bin/perl

=head1 NAME

samse2csv.pl - Reads the SAMSE output from BWA and saves a TSV file for poly-A analysis

=head1 SYNOPSYS

  % bwa samse ... | samse2csv.pl [-off -50] [-unmapped <UNMAPEDOUT>] [-fromend] [-end] > OUTPUT.TSV

=head1 DESCRIPTION

Reads the output from BWA samse mode and saves a tab delimited file to be used in poly-A analysis. This can be set to PAM reads (default using -fromend) or EJ reads (-endread with -fromend switch)


=head2 NOTES

This is still a work in progress.

=head1 COMMAND-LINE OPTIONS

Options can be abbreviated.  For example, you can use -s for
    --offset           TSV file of transcripts with length of poly A tails found in transcript annotation
    --readlength       Readlength used
    --endread          Boolean End junction (EJ) read mode
    --referencepos     Reference position is based on absolute coordinates
    --negalign         Boolean selects for negative alignments only
    --fromend          Calculates the distance of mapped reads from the end of the transcript/contig
    --quiet            Prints out extra information on tables and results
                         Default = 0


=head1 SEE ALSO

=head1 AUTHOR

Mark Okada, hmokada@sfu.ca

Copyright (c) 2009 Simon Fraser University

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.  See DISCLAIMER.txt for
disclaimers of warranty.

=cut

use strict;
use FindBin;

use lib "$FindBin::RealBin/../lib";

use Data::Dumper;
use Getopt::Long;
use Config::General;
use vars qw(%OPT %CONF);

#default vals
%OPT = (readlength => 50, quiet => 0);

GetOptions(\%OPT,
	   "unmapped=s",
	   "offset=s",
 	   "endread",
	   "negalign",
	   "referencepos=i",
           "fromend",
	   "readlength=i",
           "quiet+");


my $unmapped;
open(UNM,">$OPT{unmapped}") || die "cannot write to unmapped file" if $OPT{unmapped};
$unmapped = *UNM if $OPT{unmapped};


my @seq;
my $header;

my $offset;

# adjust the tail length by the offset file
open (OFF, $OPT{offset}) if $OPT{offset};
while (<OFF>) {
    chomp;
    my ($read,@vals) = split "\t", substr($_,1);
    $read =~ s/__.+//;
    $read =~ s/ .+//;
    $offset->{$read} = [@vals];
}
close OFF;


while (<>) {
    chomp;
    
    if (substr($_,0,1) eq ">") {
	proc_entry($header,\@seq,$unmapped, $offset) if $header;
	@seq = ();
	$header = $_;
    } else {
	push @seq, $_;
    }

    
}

if ($header) {
    proc_entry($header, \@seq,$unmapped, $offset);
}

close UNM if $OPT{unmapped};


sub proc_entry {
    my $header = shift;
    my $entries = shift;
    my $unmapped = shift;
    my $offset = shift;
    
    my $mult;
    
    
    # length of tail
    my ($tlength) = $header =~ /^\S+_(-?\d+)/;
    #print Dumper($header,$tlength);
    
    #my $length;
    #print "\n\n\n";
    #print Dumper($header,$entries);
    
    
    for (my $i=0; $i<@$entries; $i++) {
	my ($label, $dist,$n) = split("\t",$entries->[$i]);
	$dist =~ s/^\+//;
	my $strand = $dist < 0 ? "-": "+";
	
	my $length;
	my $prefix;
	my $reldist;
	
	if ($OPT{fromend}) {
	    ($prefix, $length) = $label =~ /(.*)__length=(\d+)/;
	    $label = $prefix if $prefix;
	    
	}
	
	# Rank based on MAQ mapping then relative to reference pos
	
	if ($OPT{fromend}) {
	    
	    if ($strand eq "+") {
		# SENSE
		if ($OPT{endread}) {
		    # EJ read
		    #$reldist = $dist;#t + $OPT{readlength};
		    $reldist = $dist + abs($tlength) + 1;
		} else {
		    # PAM read
		    $reldist = $length - ($dist + $OPT{readlength}) + 1;
		}
	    } else {
		# ANTISENSE - measure distance from left edge
		if ($OPT{endread}) {
		    #    BWA returns |--------------|              |---------| what we want
		    #                            EJ ****************AAAA
		    #                ========================================aaaaaaAAAAAAAAAAAAAAAAAA
		    #                               transcript               realAs    appended As
		    #
		    #                ................................................................ length
		    #                ............... -dist          ..... tlength  .................. readlength
		    #                               ..................... readlength
		    #                                                        ...... offset
		    $reldist = $length - (-$dist + $OPT{readlength} * 2) + abs($tlength) + 1;
		} else {
		    #$reldist = $length - (-$dist + $OPT{readlength}) + 1;
		    $reldist = -$dist;
		}
	    }
	    
	    
	} elsif ($OPT{referencepos}) {
	    # working with absolute distances
	    $reldist = $reldist < $OPT{referencepos} ? $OPT{referencepos} - $reldist : $reldist;
	}
	
	
	$reldist -= $offset->{$label}->[0] if $OPT{offset};
	if ($OPT{endread}) {
#	    $reldist += abs($tlength) - $OPT{readlength};
	}
	
	$mult->{$n}{$reldist}{$dist}{$label} = join("\t",$length,$strand,$tlength) if !$OPT{negalign} || $strand eq "-";

    }
    
    
    if (! $mult) {
	print $unmapped $header, "\n" if $unmapped;
	return;
    }
    
    
    # rank alignments by MAQ alignment score then by distance from end
    my $ctr = 0;
    for my $n (sort {$a<=>$b} keys %$mult) {
	for my $reldist (sort {$a<=>$b} keys %{$mult->{$n}}) {
	    for my $dist (keys %{$mult->{$n}{$reldist}}) {
		for my $label (keys %{$mult->{$n}{$reldist}{$dist}}) {
		    print join("\t",
#$absdist,
			       $header,
			       $label,
			       $dist,
			       $n,
			       $ctr++,
			       $reldist,
			       $mult->{$n}{$reldist}{$dist}{$label})."\n";
		}
	    }
	}
    }
}
