from utilities import alignment, intspan
from utilities.snv import SNV
import sys
import os
import math

fields = {1:"match",
          2:"mismatch",
          3:'repmatch',
          4:'ncount',
          5:'qnuminsert',
          6:'qbaseinsert',
          7:'tnuminsert',
          8:'tbaseinsert',
          9:"query_strand",
          10:"query",
          11:"query_len",
          12:"qstart",
          13:"qend",
          14:"target",
          15:"target_len",
          16:"tstart",
          17:"tend",
          18:"block_count",
          19:"block_sizes",
          20:"qstarts",
          21:"tstarts"
          }

def get_hsps(file, aligns=[]):
    sys.path.append('/home/rchiu/python/biopython-1.52')
    from Bio.Blast import NCBIStandalone

    result_handle = open(file, 'r')
    blast_parser = NCBIStandalone.BlastParser()
    blast_iterator = NCBIStandalone.Iterator(result_handle, blast_parser)
    sys.stderr.write("extracting hsps...\n")
    hsps = {}

    aligns_dict = {}
    if aligns:
        for a in aligns:
            aligns_dict[a.query] = {}
            aligns_dict[a.query][a.target.lstrip('chr')] = 1
        
    for blast_record in blast_iterator :
        query = blast_record.query.lstrip()

        if not aligns_dict.has_key(query):
            continue

        for align in blast_record.alignments:
            target = align.title.lstrip('>')
            
            if not aligns_dict[query].has_key(target.lstrip('chr')):
                continue

            key = query + target.lstrip('chr')
            if not hsps.has_key(key):
                hsps[key] = {}
                
            for hsp in align.hsps:
                tstart = hsp.sbjct_start
                tend = hsp.sbjct_end

                if hsp.query_end and hsp.query_start:
                    bases = abs(int(hsp.query_end) - int(hsp.query_start))
                else:
                    bases = 0
                
                hsps[key][','.join([str(tstart), str(tend)])] = [hsp.query, hsp.sbjct, hsp.match, hsp.strand, (hsp.query_start, hsp.query_end), (hsp.sbjct_start, hsp.sbjct_end), bases]
                
    sys.stderr.write("done extracting hsps\n")
    return hsps

def parse(file, filters=None, noblocks=False, noline=False, splice_motif_file=None):
    aligns = []

    prev_query = None
    prev_line = None

    group = []

    for line in open(file, 'r'):
        line = line.rstrip()

        #double lines seen in blat output?
        if line == prev_line:
            continue

        #check if line begins with number - not headers
        if line and ord(line[0]) >= ord('0') and ord(line[0]) <= ord('9'):
            cols = line.split("\t")

            #filtering
            if prev_query and not cols[9] == prev_query:
                indices = screen(group, filters)

                #only extract blocks for filtered set
                if indices:
                    for idx in indices:
                        a = create_align(group[idx], noblocks, noline)
                        aligns.append(a)
                    
                del group[:]
                
            group.append(line)

            prev_query = cols[9]
            prev_line = line

    #last group
    if group:
        indices = screen(group, filters)
        if indices:
            for idx in indices:
                a = create_align(group[idx], noblocks, noline)
                aligns.append(a)


    return aligns

def create_align(line, noblocks=False, noline=False):
    cols = line.split("\t")

    align = alignment.Alignment(method="blat")
    for field in fields:
        if fields[field] == "tstart" or fields[field] == 'qstart':
            setattr(align, fields[field], int(cols[field-1])+1)
        else:
            setattr(align, fields[field], cols[field-1])

    align.score = calc_score(cols[0], cols[1], cols[4], cols[6])
    align.match_len = int(cols[0]) + int(cols[1]) +  int(cols[2])
    align.identity = calc_identity(int(cols[11])+1, cols[12], int(cols[15])+1, cols[16], cols[4], cols[1], align.match_len)

    if not noline:
        setattr(align, 'psl_str', line)
    if not noblocks:
        get_blocks(align)

    return align


def get_blocks(align):
    align.blocks = []
    tstarts = align.tstarts.rstrip(',').split(',')
    blk_sizes = align.block_sizes.rstrip(',').split(',')
    for i in range(int(align.block_count)):
        block = [int(tstarts[i])+1, int(tstarts[i])+int(blk_sizes[i])]
        align.blocks.append(block)

    align.blocks.sort(lambda x,y: x[0]-y[0])


def screen(group, filters):
    keep = []

    scores = {}
    ranks = {}

    for idx in range(len(group)):
        cols = group[idx].split("\t")
        scores[idx] = calc_score(cols[0], cols[1], cols[4], cols[6])

    indices = scores.keys()
    indices.sort(lambda x,y: scores[y]-scores[x])

    if not filters:
        return indices

    #rank
    rank = 1
    for ii in range(len(indices)):
        if 'bestn' in filters and rank > int(filters['bestn']):
            break
        
        idx = indices[ii]
        
        if ii == 0:
            ranks[idx] = 1
        elif scores[idx] == scores[indices[ii-1]]:
            ranks[idx] = ranks[indices[ii-1]]
        else:
            rank += 1
            ranks[idx] = rank

    #skips entire group
    #unique means best alignment only has 1 hit
    if 'unique' in filters and len(indices) > 1 and scores[indices[0]] == scores[indices[1]]:
        return keep

    count = 0
    for idx in indices:
        count += 1
        cols = group[idx].split("\t")
    
        if filters.has_key('bestn') and ranks[idx] > int(filters['bestn']):
            break

        if filters.has_key('count') and count > int(filters['count']):
            break
                    
        if filters.has_key('qlen') and int(cols[10]) < int(filters['qlen']):
            break

        match_len = int(cols[0]) + int(cols[1]) +  int(cols[2])
        if filters.has_key('match'):
            if float(match_len)*100/float(cols[10]) < float(filters['match']):
                continue

        if filters.has_key('identity'):
            #100% identity, just look at number of mismatches
            if float(filters['identity']) == 100.0 and int(cols[1]) > 0:
                continue
            else: 
                identity = calc_identity(int(cols[11])+1, cols[12], int(cols[15])+1, cols[16], cols[4], cols[1], match_len)
                if identity < float(filters['identity']):
                    continue

        if filters.has_key('ungapped') and int(cols[17]) > 1:
            continue

        keep.append(idx)

    return keep

def calc_score(match, mismatch, qnuminsert, tnuminsert):
    "ucsc score"
    return int(match) - int(mismatch) - int(qnuminsert) - int(tnuminsert)

def calc_identity(qstart, qend, tstart, tend, qnuminsert, mismatch, match_len):
    "ucsc webblat identity"

    q_alisize = int(qend) - int(qstart)
    t_alisize = int(tend) - int(tstart)
    alisize = min(q_alisize, t_alisize)

    if alisize <= 0:
        return 0

    sizediff = q_alisize - t_alisize
    if sizediff < 0:
        sizediff = 0

    insert_factor = int(qnuminsert)

    millibad = (1000 * (float(mismatch) + insert_factor + round(3 * math.log(1 + sizediff)))) / float(match_len)

    return round(100.0 - millibad * 0.1, 1)

def get_snvs(align, hsps, hsps_used):
    sf = alignment.SNVfinder(align)
    all_snvs = []

    #hopefullly, one block = one hsp
    count_hsp = 0
    for block in align.blocks:
        found = False
        key = align.query + align.target.lstrip('chr')
        if hsps.has_key(key):
            #sort spans to have biggest first checked first
            spans = hsps[key].keys()
            spans.sort(lambda x,y: int(hsps[key][y][-1])- int(hsps[key][x][-1]))

            span_matched = None

            #try to find exact match first
            for span in spans:
                tcoords = [hsps[key][span][5][0], hsps[key][span][5][1]]
                tcoords.sort(lambda x,y: int(x)-int(y))

                if int(tcoords[0]) == int(block[0]) and int(tcoords[1]) == int(block[1]):
                    #print "found exact:", align.query, block, tcoords
                    span_matched = span

            #if no exact match to both ends, pick the one that matches one end
            if not span_matched:
                for span in spans:
                    span1, span2 = span.split(',')
                    if int(span1) == int(block[0]) or int(span2) == int(block[1]):
                        #print "found olap1:", align.query, block, [hsps[key][span][5][0], hsps[key][span][5][1]]
                        span_matched = span
                        break
                    
            #if not, find the biggest that overlaps
            if not span_matched:
                for span in spans:
                    if intspan.overlap(span.split(','), block):
                        #print "found olap2:", align.query, block, [hsps[key][span][5][0], hsps[key][span][5][1]]
                        span_matched = span
                        break

            if span_matched:
                found = True
                count_hsp += 1

                if not hsps_used.has_key(key+span_matched):
                    hsps_used[key+span_matched] = 1
                                    
                    hsp = hsps[key][span_matched]

                    query_strand = True
                    if not hsp[3][0].lower() == "plus":
                        query_strand = False
                    target_strand = True
                    if not hsp[3][1].lower() == "plus":
                        target_strand = False

                    snvs = sf.find_snv([hsp[0],hsp[1]], hsp[2], [hsp[4][0], hsp[5][0]], query_strand, target_strand, align.method)
                    if snvs:
                        all_snvs.extend(snvs)

        if not found:
            sys.stderr.write("no hsp: %s %s %s-%s\n" % (align.query, align.target, block[0], block[1]))

    return all_snvs, count_hsp

if __name__ == '__main__':
    parse(sys.argv[1])
