from align_parser import parse
import assembly
import track
from optparse import OptionParser
import re
import sys
import glob, os
import hashlib

class ContigMerger:

    align_file = None
    aligner = None
    query_k = None
    target_k = None
    query_fa = None
    target_fa = None
    buried = {}
    num_buried = 0
    log_file = None
    log = None
    reverse_align_file = None
    first_round = None

    def __init__(self, query_fa, target_fa, align_file, aligner, log_file=None, reverse=None, first_round=False):
        self.align_file = align_file
        self.aligner = aligner
        self.query_fa = query_fa
        self.target_fa = target_fa
        self.reverse_align_file = reverse
        self.first_round = first_round
        
        self.log_file = log_file
        if self.log_file:
            self.log = open(self.log_file, 'w')
        
        m = re.search('(k\d+).*(k\d+)', align_file)
        if m and m.group:
            (self.query_k, self.target_k) = (m.group(1), m.group(2))

        if self.log:
            self.log.write("query k-value=%s\ntarget k-value=%s\n" % (self.query_k, self.target_k))

    def bury(self):
        #parse alignments
        sys.stderr.write('parsing alignments\n')
        filters = {'identity':100.0, 'match':100.0, 'ungapped':True};
        
        aligns = parse(self.align_file, self.aligner, filters)
        if aligns:
            if not aligns[0].query[0] == 'k':
                [self.attach_k(a, self.query_k, query=True) for a in aligns]
            if not aligns[0].target[0] == 'k':
                [self.attach_k(a, self.target_k, target=True) for a in aligns]

        if self.reverse_align_file:
            reverse_aligns = parse(self.reverse_align_file, self.aligner, filters)
            if reverse_aligns:
                if not reverse_aligns[0].query[0] == 'k':
                    [self.attach_k(a, self.target_k, query=True) for a in reverse_aligns]
                if not reverse_aligns[0].target[0] == 'k':
                    [self.attach_k(a, self.query_k, target=True) for a in reverse_aligns]
                aligns.extend(reverse_aligns)        

        sys.stderr.write('burying contigs\n')
        for a in aligns:
            contig_p = re.compile('^(\S+)')
            m = contig_p.search(a.query)
            query_contig = m.group(1)
            m = contig_p.search(a.target)
            target_contig = m.group(1)

            if query_contig == target_contig:
                continue

            child = None
            parent = None
            identical = False
            
            #target should be larger k, always check if larger k can be buried in smaller k first
            #identical contigs
            if int(a.query_len) == int(a.target_len):
                mq = re.search('k(\d+)', query_contig)
                mt = re.search('k(\d+)', target_contig)
                identical = True

                if mq and mt:
                    #query is lower k
                    if int(mq.group(1)) < int(mt.group(1)):
                        child = target_contig
                        parent = query_contig
                    else:
                        child = query_contig
                        parent = target_contig
                    
            elif int(a.match_len) == int(a.target_len):
                child = target_contig
                parent = query_contig
                    
            elif int(a.match_len) == int(a.query_len):
                child = query_contig
                parent = target_contig

            if child and parent:
                if not self.buried.has_key(child):
                    self.buried[child] = [parent]
                else:
                    self.buried[child].append(parent)
                #self.buried[child] = parent
                if self.log:
                    identical_str = ""
                    if identical:
                        identical_str = " (identical)"
                    self.log.write("%s buried in %s%s\n" % (child, parent, identical_str))

        self.num_buried = len(set(self.buried.keys()))
        if self.log:
            self.log.write("num contigs buried: %s\n" % self.num_buried)

    def attach_k2(self, a, k_query, k_target):
        "attach k-value to query and target"

        contig_p = re.compile('^(\S+)')
        m = contig_p.search(a.query)
        if m:
            a.query = ":".join((k_query, m.group(1)))

        m = contig_p.search(a.target)
        if m:
            a.target = ":".join((k_target, m.group(1)))

    def attach_k(self, a, k, target=False, query=False):
        "attach k-value to query and target"

        contig_p = re.compile('^(\S+)')

        if target:
            m = contig_p.search(a.target)
            a.target = ":".join((k, m.group(1)))
        else:
            m = contig_p.search(a.query)
            a.query = ":".join((k, m.group(1)))

    def merge(self, output_file=None):
        lines = []

        sys.stderr.write('getting query contigs\n')
        kq = assembly.Assembly(self.query_k)
        kq.fasta = self.query_fa
        q_contigs = kq.get_contigs(sequence=True)
        
        #mapping contig number to contig object
        if not 'k' in q_contigs[0].num:
            contig_dict = dict((self.query_k + ":" + str(c.num), c) for c in q_contigs)
        else:
            contig_dict = dict((str(c.num), c) for c in q_contigs)

        q_contigs_dict = contig_dict.copy()

        sys.stderr.write('getting target contigs\n')
        kt = assembly.Assembly(self.target_k)
        kt.fasta = self.target_fa
        t_contigs = kt.get_contigs(sequence=True)
        if not 'k' in t_contigs[0].num:
            contig_dict.update(dict((self.target_k + ":" + str(c.num), c) for c in t_contigs))
        else:
            contig_dict.update(dict((str(c.num), c) for c in t_contigs))

        #canonical list
        canon_ctg_nums = list(set(contig_dict.keys()) - set(self.buried.keys()))

        #untouched
        parents = []
        for p in self.buried.values():
            parents.extend(p)
            
        untouched = list(set(canon_ctg_nums) - set(parents))
        untouched_dict = dict((c,1) for c in untouched)

        if self.log:
            self.log.write("%d(buried) + %d(canonical) = %d\n" % (len(self.buried.keys()), len(canon_ctg_nums), len(self.buried.keys())+len(canon_ctg_nums)))
            self.log.write("%d(query) + %d(target) = %d\n" % (len(q_contigs), len(t_contigs), len(q_contigs)+len(t_contigs)))

        #write out contigs
        sys.stderr.write('screening and writing out contigs\n')
        for n in sorted(canon_ctg_nums):
            ctg = contig_dict[n]
            ctg.num = n

            #indicate untouched contigs, only in first round
            #there should be no more untouched contigs in later rounds
            if untouched_dict.has_key(n):
                if self.first_round and not n[-1] == 'u':
                    if self.first_round == 'b':
                        ctg.num += 'u'
                    elif self.first_round == 'q' and q_contigs_dict.has_key(n):
                        ctg.num += 'u'
                    elif self.first_round == 't' and not q_contigs_dict.has_key(n):
                        ctg.num += 'u' 
            #parent
            elif n[-1] == 'u':
                #print ctg.num, " is parent now"
                ctg.num = ctg.num.rstrip('u')
                
            lines.append(ctg.fasta())

        out = open(output_file, 'w')
        out.writelines(lines)
        out.close()

    def copy_track(self, fa_file, tracks_dir, output_track, track_name):
        ass = assembly.Assembly(k=0)
        ass.fasta = fa_file
        contigs = ass.get_contigs()
        k_contigs = {}
        for c in contigs:
            k, ctg = c.num.split(':')
            if not k_contigs.has_key(k):
                k_contigs[k] = {}                
            k_contigs[k][ctg] = 0

        align_files = {}
        for align_file in glob.glob(os.path.join(tracks_dir,'*k[0-9]*')):
            m = re.search('(k\d+)', align_file)
            k = None
            if m and m.group(1):
                k = m.group(1)

            if k:
                align_files[k] = align_file

        #all the alignments and contigs with no alignments
        all_aligns = []
        no_aligns = []
        for k in sorted(k_contigs.keys()):
            #no alignment file for k
            if not align_files.has_key(k):
                no_aligns.extend([":".join((k,c)) for c in k_contigs[k].keys()])
                if self.log:
                    self.log.write("alignments %s\t%d\t%d\t%d\n" % (k, len(k_contigs[k].keys()), 0, len(k_contigs[k].keys())))
                continue

            align_file = align_files[k]
            ext = os.path.splitext(align_file)[1]

            k_aligns = []
            if ext == '.gff':
                trk = track.Track(align_file)
                aligns = trk.parse_gff()

                for a in aligns:
                    if k_contigs[k].has_key(a.query):
                        k_contigs[k][a.query] += 1
                        k_aligns.append(a)
                        
            na = [c for c in k_contigs[k].keys() if k_contigs[k][c] == 0]
            no_aligns.extend(na)

            self.log.write("alignments %s\t%d\t%d\t%d\n" % (k, len(k_contigs[k].keys()), len(k_aligns), len(na)))

            for a in k_aligns:
                 a.query = k + ":" + a.query
            all_aligns.extend(k_aligns)

        if self.log:
            self.log.write("total contigs\t%d\n" % (len(contigs)))
            self.log.write("no alignments:\t%d\n" % (len(no_aligns)))
            self.log.write("alignments output:\t%d\n" % (len(all_aligns)))
        
        if all_aligns:
            trk = track.Track()
            trk.name = track_name
            trk.desc = trk.name
            trk.output(all_aligns, output_track)

def batch(topdir):
    pass

if __name__ == '__main__':
    usage = "Usage: %prog k1.fa k2.fa k1_k2_align aligner out.fa"

    parser = OptionParser(usage=usage)
    parser.add_option("-1", "--first", dest="first", help="first round of merging: b=both, q=query, t=target")
    parser.add_option("-r", "--reverse", dest="reverse", help="reverse alignment file")
    parser.add_option("-t", "--track", dest="track", help="input_track_dir output_track_file output_track_name", nargs=3)
    parser.add_option("-l", "--log", dest="log", help="log file")

    (options, args) = parser.parse_args()

    if len(args) == 5:
        cm = ContigMerger(args[0], args[1], args[2], args[3], log_file=options.log, reverse=options.reverse, first_round=options.first)
        cm.bury()
        output_file = cm.merge(args[4])

        if options.track:
            cm.copy_track(args[4], options.track[0], options.track[1], options.track[2])

        if cm.log:
            cm.log.close()
    else:
        parser.error("incorrect number of arguments")
